package com.ur.urcap.externalaxis.axisgroupjogging.toolbar;

import com.ur.urcap.api.contribution.toolbar.ToolbarAPIProvider;
import com.ur.urcap.api.contribution.toolbar.ToolbarContext;
import com.ur.urcap.api.contribution.toolbar.swing.SwingToolbarContribution;
import com.ur.urcap.api.domain.userinteraction.inputvalidation.InputValidationFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardNumberInput;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardTextInput;
import com.ur.urcap.externalaxis.axisgroupjogging.impl.Style;
import com.ur.urcap.externalaxis.axisgroupjogging.impl.V5Style;
import com.ur.urcap.externalaxis.axisgroupjogging.rtde.RTDEClient;
import com.ur.urcap.externalaxis.axisgroupjogging.rtde.RTDEOutputHandler;
import com.ur.urcap.externalaxis.axisgroupjogging.rtde.RTDEOutputParam;
import com.ur.urcap.externalaxis.axisgroupjogging.URCommunicator.ScriptCommand;
import com.ur.urcap.externalaxis.axisgroupjogging.URCommunicator.ScriptSender;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.stream.Collectors;

class AxisGroupJoggingToolbarContribution implements SwingToolbarContribution {

	private final ToolbarAPIProvider apiProvider;
	private final InputValidationFactory validatorFactory;
	private final KeyboardInputFactory keyboardInputFactory;

	// GUI
	private static final int MAXIMUM_FRACTION_DIGITS = 3;
	private final Style style;
	private ImageIcon axisGroupJogNegativeIcon;
	private ImageIcon axisGroupJogPositiveIcon;
	private ImageIcon axisGroupJogActiveIcon;
	private ImageIcon axisGroupJogInactiveIcon;
	private JButton axisGroupJogNegativeButton = new JButton();
	private JButton axisGroupJogPositiveButton = new JButton();
	private JLabel axisGroupJogStatus = new JLabel();
	private JTextField axisGroupName = new JTextField();
	private JTextField axisGroupMaxVelocity = new JTextField("0.0");
	private JTextField axisGroupAccelerationFactor = new JTextField("0.0");
	private JLabel axisGroupActualPositions = new JLabel("[0.0, 0.0, 0.0, 0.0, 0.0, 0.0");
	private JLabel axisGroupTargetPositions = new JLabel("[0.0, 0.0, 0.0, 0.0, 0.0, 0.0");
	private JTextArea axisGroupActiveCommand = new JTextArea("");
	private DecimalFormat df = new DecimalFormat("#");

	// ScriptCommunicator
	private final ScriptSender sender;

	// RTDE
	private final RTDEClient client;

	AxisGroupJoggingToolbarContribution(ToolbarContext context) {
		apiProvider = context.getAPIProvider();
		style = new V5Style();
		keyboardInputFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getKeyboardInputFactory();
		validatorFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getInputValidationFactory();

		sender = new ScriptSender();

		client = new RTDEClient(new RTDEOutputHandler() {
			@Override
			public void onProtocolVersion(int protocolVersion) {
				if (protocolVersion != 2) {
					throw new RuntimeException("Protocol v2 expected");
				}
			}

			@Override
			public void onData(RTDEOutputParam output, Object value) {
				switch (output) {
					case actual_external_q: {
						double[] data = (double[]) value;
						String stringData = Arrays.stream(data)
								.mapToObj(df::format)
								.collect(Collectors.joining(", "));
						axisGroupActualPositions.setText(stringData);
						break;
					}
					case target_external_q: {
						double[] data = (double[]) value;
						String stringData = Arrays.stream(data)
								.mapToObj(df::format)
								.collect(Collectors.joining(", "));
						axisGroupTargetPositions.setText(stringData);
						break;
					}
					default:
						break;
				}
			}
		});

		df.setMaximumFractionDigits(MAXIMUM_FRACTION_DIGITS);
	}

	@Override
	public void openView() {
		sender.connect("127.0.0.1", ScriptSender.PRIMARY_PORT, 200);
		if (client.connect("127.0.0.1", RTDEClient.PORT, 200)) {
			try {
				client.requestProtocolVersion(2);
				client.setupOutputsV2(50,
						RTDEOutputParam.actual_external_q,
						RTDEOutputParam.target_external_q);
				client.startOutput();
			} catch (Exception e) {
				System.out.println("RTDE client setup failed");
				System.out.println(e);
			}
		}
	}

	@Override
	public void closeView() {
		sender.disconnect();
		client.disconnect();
	}

	public void buildUI(JPanel jPanel) {
		axisGroupJogNegativeIcon = new ImageIcon(getClass().getResource("/icons/left_arrow3.png"));
		axisGroupJogPositiveIcon = new ImageIcon(getClass().getResource("/icons/right_arrow3.png"));
		axisGroupJogActiveIcon = new ImageIcon(getClass().getResource("/icons/G5_led_green.png"));
		axisGroupJogInactiveIcon = new ImageIcon(getClass().getResource("/icons/G5_led_off.png"));

		JLabel axisGroupLabel = new JLabel("Axis Group:");
		setupTextInputField(axisGroupName, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForAxisGroupName();
				keyboard.show(axisGroupName, getCallbackForAxisGroupName());
			}
		});

		JLabel axisGroupMaxVelocityLabel = new JLabel("Max Velocity:");
		setupTextInputField(axisGroupMaxVelocity, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardNumberInput<Double> keyboard = getKeyboardForAxisGroupMaxVelocity();
				keyboard.show(axisGroupMaxVelocity, getCallbackForAxisGroupMaxVelocity());
			}
		});

		JLabel axisGroupAccelerationFactorLabel = new JLabel("Acceleration Factor:");
		setupTextInputField(axisGroupAccelerationFactor, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardNumberInput<Double> keyboard = getKeyboardForAxisGroupAccelerationFactor();
				keyboard.show(axisGroupAccelerationFactor, getCallbackForAxisGroupAccelerationFactor());
			}
		});

		JLabel axisGroupJogLabel = new JLabel("Jog:");
		setupAxisGroupJogControl(
				axisGroupJogNegativeButton,
				axisGroupJogNegativeIcon,
				new MouseAdapter() {
					@Override
					public void mousePressed(MouseEvent e) {
						startJogAxisGroupNegative();
					}

					@Override
					public void mouseReleased(MouseEvent e) {
						stopAxisGroup();
					}
				});

		setupAxisGroupJogControl(
				axisGroupJogPositiveButton,
				axisGroupJogPositiveIcon,
				new MouseAdapter() {
					@Override
					public void mousePressed(MouseEvent e) {
						startJogAxisGroupPositive();
					}

					@Override
					public void mouseReleased(MouseEvent e) {
						stopAxisGroup();
					}
				});

		setupAxisGroupJogStatus(axisGroupJogStatus, axisGroupJogInactiveIcon);

		JLabel axisGroupActualPositionsLabel = new JLabel("actual_external_q:");
		setupDisplayField(axisGroupActualPositions);

		JLabel axisGroupTargetPositionsLabel = new JLabel("target_external_q:");
		setupDisplayField(axisGroupTargetPositions);

		JLabel axisGroupActiveCommandLabel = new JLabel("Sent Command:");
		setupDebugField(axisGroupActiveCommand);

		jPanel.setLayout(new BoxLayout(jPanel, BoxLayout.Y_AXIS));

		JPanel axisGroupPanel = new JPanel();
		GroupLayout layout = new GroupLayout(axisGroupPanel);
		axisGroupPanel.setLayout(layout);
		layout.setAutoCreateGaps(true);
		layout.setAutoCreateContainerGaps(true);

		layout.setHorizontalGroup(
				layout.createSequentialGroup()
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
								.addComponent(axisGroupLabel)
								.addComponent(axisGroupMaxVelocityLabel)
								.addComponent(axisGroupAccelerationFactorLabel)
								.addComponent(axisGroupJogLabel)
						)
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
								.addComponent(axisGroupName)
								.addComponent(axisGroupMaxVelocity)
								.addComponent(axisGroupAccelerationFactor)
								.addGroup(layout.createSequentialGroup()
										.addComponent(axisGroupJogNegativeButton)
										.addComponent(axisGroupJogPositiveButton)
										.addComponent(axisGroupJogStatus)
								)
						)
		);

		layout.setVerticalGroup(
				layout.createSequentialGroup()
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupLabel)
								.addComponent(axisGroupName)
						)
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupMaxVelocityLabel)
								.addComponent(axisGroupMaxVelocity)
						)
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupAccelerationFactorLabel)
								.addComponent(axisGroupAccelerationFactor)
						)
						.addGroup(layout.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupJogLabel)
								.addComponent(axisGroupJogNegativeButton)
								.addComponent(axisGroupJogPositiveButton)
								.addComponent(axisGroupJogStatus)
						)
		);

		JPanel rtdePanel = new JPanel();
		GroupLayout rtdeLayot = new GroupLayout(rtdePanel);
		rtdePanel.setLayout(rtdeLayot);
		rtdePanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "RTDE Data"));
		rtdeLayot.setAutoCreateGaps(true);
		rtdeLayot.setAutoCreateContainerGaps(true);

		rtdeLayot.setHorizontalGroup(
				rtdeLayot.createSequentialGroup()
						.addGroup(rtdeLayot.createParallelGroup(GroupLayout.Alignment.LEADING)
								.addComponent(axisGroupActualPositionsLabel)
								.addComponent(axisGroupTargetPositionsLabel)
						)
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, style.getExtraLargeVerticalSpacing(), Short.MAX_VALUE)
						.addGroup(rtdeLayot.createParallelGroup(GroupLayout.Alignment.LEADING)
								.addComponent(axisGroupActualPositions)
								.addComponent(axisGroupTargetPositions)
						)
		);

		rtdeLayot.setVerticalGroup(
				rtdeLayot.createSequentialGroup()
						.addGroup(rtdeLayot.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupActualPositionsLabel)
								.addComponent(axisGroupActualPositions)
						)
						.addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, style.getVerticalSpacing(), Short.MAX_VALUE)
						.addGroup(rtdeLayot.createParallelGroup(GroupLayout.Alignment.BASELINE)
								.addComponent(axisGroupTargetPositionsLabel)

								.addComponent(axisGroupTargetPositions)
						)
		);

		JPanel commandPanel = new JPanel();
		GroupLayout commandLayout = new GroupLayout(commandPanel);
		commandPanel.setLayout(commandLayout);
		rtdeLayot.setAutoCreateGaps(true);
		rtdeLayot.setAutoCreateContainerGaps(true);

		commandLayout.setHorizontalGroup(
				commandLayout.createParallelGroup(GroupLayout.Alignment.LEADING)
						.addComponent(axisGroupActiveCommandLabel)
						.addComponent(axisGroupActiveCommand)
		);

		commandLayout.setVerticalGroup(
				commandLayout.createSequentialGroup()
						.addComponent(axisGroupActiveCommandLabel)
						.addComponent(axisGroupActiveCommand)
		);

		axisGroupPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		axisGroupPanel.setAlignmentY(Component.TOP_ALIGNMENT);
		rtdePanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		rtdePanel.setAlignmentY(Component.TOP_ALIGNMENT);
		commandPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		commandPanel.setAlignmentY(Component.TOP_ALIGNMENT);
		jPanel.add(axisGroupPanel);
		jPanel.add(rtdePanel);
		jPanel.add(commandPanel);
	}

	private void setupTextInputField(final JTextField inputField, MouseAdapter mouseAdapter) {
		inputField.setFocusable(false);
		inputField.setPreferredSize(style.getInputfieldSize());
		inputField.setMaximumSize(inputField.getPreferredSize());
		inputField.addMouseListener(mouseAdapter);
	}

	private KeyboardTextInput getKeyboardForAxisGroupName() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(axisGroupName.getText());
		keyboardInput.setErrorValidator(validatorFactory.createStringLengthValidator(1, 255));
		return keyboardInput;
	}

	private KeyboardInputCallback<String> getCallbackForAxisGroupName() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				axisGroupName.setText(value);
			}
		};
	}

	private KeyboardNumberInput<Double> getKeyboardForAxisGroupMaxVelocity() {
		KeyboardNumberInput<Double> keyboardInput = keyboardInputFactory.createPositiveDoubleKeypadInput();
		keyboardInput.setInitialValue(Double.parseDouble(axisGroupMaxVelocity.getText()));
		return keyboardInput;
	}

	private KeyboardInputCallback<Double> getCallbackForAxisGroupMaxVelocity() {
		return new KeyboardInputCallback<Double>() {
			@Override
			public void onOk(Double value) {
				String stringValue = df.format((double) value);
				axisGroupMaxVelocity.setText(stringValue);
			}
		};
	}

	private KeyboardNumberInput<Double> getKeyboardForAxisGroupAccelerationFactor() {
		KeyboardNumberInput<Double> keyboardInput = keyboardInputFactory.createPositiveDoubleKeypadInput();
		keyboardInput.setInitialValue(Double.parseDouble(axisGroupAccelerationFactor.getText()));
		keyboardInput.setErrorValidator(validatorFactory.createDoubleRangeValidator(0.0, 1.0));
		return keyboardInput;
	}

	private KeyboardInputCallback<Double> getCallbackForAxisGroupAccelerationFactor() {
		return new KeyboardInputCallback<Double>() {
			@Override
			public void onOk(Double value) {
				String stringValue = df.format((double) value);
				axisGroupAccelerationFactor.setText(stringValue);
			}
		};
	}

	private void setupAxisGroupJogControl(
			final JButton jogButton,
			final ImageIcon jogIcon,
			MouseAdapter jogMouseAdapter) {

		jogButton.setIcon(jogIcon);
		jogButton.setFocusable(false);
		jogButton.setPreferredSize(style.getButtonSize());
		jogButton.setMaximumSize(jogButton.getPreferredSize());
		jogButton.addMouseListener(jogMouseAdapter);
	}

	private void setupAxisGroupJogStatus(
			final JLabel jogStatusLabel,
			final ImageIcon jogInactiveIcon) {

		jogStatusLabel.setIcon(jogInactiveIcon);
		jogStatusLabel.setFocusable(false);
		jogStatusLabel.setPreferredSize(style.getButtonSize());
		jogStatusLabel.setMaximumSize(jogStatusLabel.getPreferredSize());
	}

	private void startJogAxisGroupNegative() {
		ScriptCommand cmd = new ScriptCommand("urcap_external_axis_jog_negative");
		cmd.appendLine(String.format(
				"axis_group_speedj(\"%s\", [-%s], %s, 100)",
				axisGroupName.getText(),
				axisGroupMaxVelocity.getText(),
				axisGroupAccelerationFactor.getText()));
		if (sender.executeURScript(cmd)) {
			axisGroupJogStatus.setIcon(axisGroupJogActiveIcon);
			axisGroupActiveCommand.setText(cmd.toString());
		}

	}

	private void startJogAxisGroupPositive() {
		ScriptCommand cmd = new ScriptCommand("urcap_external_axis_jog_positive");
		cmd.appendLine(String.format(
				"axis_group_speedj(\"%s\", [%s], %s, 100)",
				axisGroupName.getText(),
				axisGroupMaxVelocity.getText(),
				axisGroupAccelerationFactor.getText()));
		if (sender.executeURScript(cmd)) {
			axisGroupJogStatus.setIcon(axisGroupJogActiveIcon);
			axisGroupActiveCommand.setText(cmd.toString());
		}
	}

	private void stopAxisGroup() {
		ScriptCommand cmd = new ScriptCommand("urcap_external_axis_jog_stop");
		cmd.appendLine(String.format(
				"axis_group_speedj(\"%s\", [0], %s)",
				axisGroupName.getText(),
				axisGroupAccelerationFactor.getText()));
		if (sender.executeURScript(cmd)) {
			axisGroupJogStatus.setIcon(axisGroupJogInactiveIcon);
			axisGroupActiveCommand.setText(cmd.toString());
		}
	}

	private void setupDisplayField(final JLabel displayLabel) {
		displayLabel.setFocusable(false);
		displayLabel.setPreferredSize(style.getDisplayfieldSize());
		displayLabel.setMaximumSize(displayLabel.getPreferredSize());
	}

	private void setupDebugField(final JTextArea debugArea) {
		debugArea.setFocusable(false);
		debugArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, style.getDebugFontSize()));
	}
}
