package com.ur.urcap.externalaxis.coordinatedmotionprogrammer.impl;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.ObjectInputStream;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;

import com.ur.urcap.api.contribution.ProgramNodeContribution;
import com.ur.urcap.api.contribution.program.ProgramAPIProvider;
import com.ur.urcap.api.domain.data.DataModel;
import com.ur.urcap.api.domain.script.ScriptWriter;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardNumberInput;
import com.ur.urcap.api.domain.undoredo.UndoableChanges;

// This class contains the implementation details for a
// MoveP/MoveCWithAxisGroup node that has robot and axis waypoints.  It
// contains tools that retrieve the current axis and robot poses and store them
// as waypoints for manual wyapoint teaching.
public
class BaseMoveWithAxisGroupNodeContribution extends
    BaseProcessMoveNodeContribution implements BaseProcessMoveNodeInterface {
 private
  BaseMoveWithAxisGroupNodeView view;

 private
  static final String AXIS_WAYPOINT_KEY =
      "AXIS_WAYPOINT";  // maps the axis waypoint in the data model stored
                        // as
                        // a string that represents a URScript double
                        // array,
                        // e.g.,
                        // "[0.123, 0.124]"

 private
  static final String ROBOT_WAYPOINT_KEY =
      "ROBOT_WAYPOINT";  // maps the robot waypoint in the data model
                         // stored
                         // as the TCP pose expressed in the tracked frame
                         // and
                         // represented as a URScript pose, e.g.,
                         // "p[0.1,0.1,0.1,0.1,0.1,0.1]"

 private
  final DataModel model;

 private
  void setAxisWaypoint(final String axisWaypoint) {
    view.setAxisWaypointInputField(axisWaypoint);

    undoRedoManager.recordChanges(new UndoableChanges() {
      @Override public void executeChanges() {
        model.set(AXIS_WAYPOINT_KEY, axisWaypoint);
      }
    });
  }

 private
  void setRobotWaypoint(final String robotWaypoint) {
    view.setRobotWaypointInputField(robotWaypoint);

    undoRedoManager.recordChanges(new UndoableChanges() {
      @Override public void executeChanges() {
        model.set(ROBOT_WAYPOINT_KEY, robotWaypoint);
      }
    });
  }

 public
  BaseMoveWithAxisGroupNodeContribution(ProgramAPIProvider apiProvider,
                                        BaseMoveWithAxisGroupNodeView view,
                                        DataModel model) {
    super(apiProvider, model, view);

    this.view = view;
    this.model = model;
  }

 public
  void openView() {
    super.openView();
    view.setAxisWaypointInputField(getAxisWaypoint());
    view.setRobotWaypointInputField(getRobotWaypoint());
  }

  // Indicates that the node is defined when axis group, axis waypoint, and
  // robot waypoints have been defined.
 public
  boolean isDefined() {
    return !(getAxisGroup().isEmpty() || getAxisWaypoint().isEmpty() ||
             getRobotWaypoint().isEmpty());
  }

 public
  String getAxisWaypoint() { return model.get(AXIS_WAYPOINT_KEY, ""); }

 public
  String getRobotWaypoint() { return model.get(ROBOT_WAYPOINT_KEY, ""); }

 public
  KeyboardInputCallback<String> getCallbackForAxisWaypoint() {
    return new KeyboardInputCallback<String>() {
      @Override public void onOk(final String value) {
        setAxisWaypoint(value);
      }
    };
  }

 public
  KeyboardInputCallback<String> getCallbackForRobotWaypoint() {
    return new KeyboardInputCallback<String>() {
      @Override public void onOk(final String value) {
        setRobotWaypoint(value);
      }
    };
  }

  // This builds the program for capturing the robot waypoint that will be sent
  // to the robot over the primary interface. The program sends the waypoint
  // back to this java node over a socket. The waypoint is the tcp pose in the
  // tracked frame.
 public
  String getRobotWaypointCaptureProgram() {
    // Construct the secondary program to be sent to the controller
    StringBuilder programBuilder = new StringBuilder();

    programBuilder.append("sec mySecondaryProgram():\n");
    programBuilder.append("  tcpInTrackedFrame = get_pose(\"tcp\", \"" +
                          getTrackedFrame() + "\")\n");
    programBuilder.append("  socket_open(\"127.0.0.1\", " +
                          SOCKET_PORT.toString() + ", \"my_socket\")\n");
    programBuilder.append(
        "  socket_send_line( tcpInTrackedFrame, \"my_socket\")\n");
    programBuilder.append("  socket_close(\"my_socket\")\n");

    programBuilder.append("end\n");

    return programBuilder.toString();
  }

  // This builds the program for capturing the axis waypoint that will be sent
  // to the robot over the primary interface. The program sends the waypoint
  // back to this java node over a socket.
 public
  String getAxisWaypointCaptureProgram() {
    StringBuilder programBuilder = new StringBuilder();

    programBuilder.append("sec mySecondaryProgram():\n");
    programBuilder.append(
        "  currentTargets = axis_group_get_target_positions(\"" +
        getAxisGroup() + "\")\n");
    programBuilder.append("  socket_open(\"127.0.0.1\", " +
                          SOCKET_PORT.toString() + ", \"my_socket\")\n");
    programBuilder.append(
        "  socket_send_line( currentTargets, \"my_socket\")\n");
    programBuilder.append("  socket_close(\"my_socket\")\n");
    programBuilder.append("end\n");

    return programBuilder.toString();
  }

  // Gets the TCP pose in the tracked frame from the robot and stores it in the
  // DataModel as the robot waypoint.
 public
  void teachRobotWaypoint() {
    final String robotWaypoint =
        getStringResponseBackFromRobot(getRobotWaypointCaptureProgram());

    setRobotWaypoint(robotWaypoint);
  }

  // Gets the axis position from the robot and stores it in the DataModel as
  // the axis waypoint.
 public
  void teachAxisWaypoint() {
    final String axisWaypoint =
        getStringResponseBackFromRobot(getAxisWaypointCaptureProgram());

    setAxisWaypoint(axisWaypoint);
  }
}
