package com.ur.urcap.externalaxis.coordinatedmotionprogrammer.impl;

import java.awt.Font;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import com.ur.urcap.api.contribution.ContributionProvider;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardNumberInput;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardTextInput;

// This class implements a GUI for a generic program node that contains speed,
// acceleration, and blend radius.
public class BaseProcessMoveNodeView {

  protected final Style style;

  protected KeyboardInputFactory keyboardInputFactory;

  private JTextField speedInputField = new JTextField();

  private JTextField accelerationInputField = new JTextField();

  private JTextField blendRadiusInputField = new JTextField();

  public Component createHorizontalSpacing() {
    return Box.createRigidArea(new Dimension(style.getHorizontalSpacing(), 0));
  }

  public Component createVerticalSpacing() {
    return createVerticalSpacing(style.getVerticalSpacing());
  }

  public Component createVerticalSpacing(int space) {
    return Box.createRigidArea(new Dimension(0, space));
  }

  // Gets a keyboard for inputting positive doubles.
  public KeyboardNumberInput<Double> getKeyboard(final double initialValue) {
    KeyboardNumberInput<Double> keyboard = keyboardInputFactory.createPositiveDoubleKeypadInput();

    keyboard.setInitialValue(initialValue);

    return keyboard;
  }

  // Gets a keyboard for inputting text.
  public KeyboardTextInput getKeyboard(final String initialValue) {
    KeyboardTextInput keyboard = keyboardInputFactory.createStringKeyboardInput();

    keyboard.setInitialValue(initialValue);

    return keyboard;
  }

  // Creates a text input field with a label.
  public Box createLabelInputField(final String label, final JTextField inputField,
      MouseAdapter mouseAdapter) {
    Box horizontalBox = Box.createHorizontalBox();
    horizontalBox.setAlignmentX(Component.LEFT_ALIGNMENT);

    JLabel jLabel = new JLabel(label);

    inputField.setFocusable(false);
    inputField.setPreferredSize(style.getInputfieldSize());
    inputField.setMaximumSize(inputField.getPreferredSize());
    inputField.addMouseListener(mouseAdapter);

    horizontalBox.add(jLabel);
    horizontalBox.add(createHorizontalSpacing());
    horizontalBox.add(inputField);

    return horizontalBox;
  }

  public BaseProcessMoveNodeView(final Style style,
      KeyboardInputFactory keyboardInputFactory) {
    this.style = style;
    this.keyboardInputFactory = keyboardInputFactory;
  }

  public void setSpeedInputField(final Double speed) {
    speedInputField.setText(speed.toString());
  }

  public void setAccelerationInputField(final Double acceleration) {
    accelerationInputField.setText(acceleration.toString());
  }

  public void setBlendRadiusInputField(final Double blendRadius) {
    blendRadiusInputField.setText(blendRadius.toString());
  }

  // Creates a panel for users to input speed.
  public <T extends BaseProcessMoveNodeContribution> Component createSpeedPanel(
      final ContributionProvider<T> provider) {

    return createLabelInputField(
        "Speed (m/s):", speedInputField, new MouseAdapter() {
          @Override
          public void mousePressed(MouseEvent e) {
            if (speedInputField.isEditable()) {
              KeyboardNumberInput<Double> keyboard = getKeyboard(provider.get().getSpeed());
              keyboard.show(speedInputField,
                  provider.get().getCallbackForSpeed());
            }
          }
        });
  }

  // Creates a panel for users to input acceleration.
  public <T extends BaseProcessMoveNodeContribution> Component createAccelerationPanel(
      final ContributionProvider<T> provider) {

    return createLabelInputField(
        "Acceleration (m/s/s):", accelerationInputField, new MouseAdapter() {
          @Override
          public void mousePressed(MouseEvent e) {
            if (accelerationInputField.isEditable()) {
              KeyboardNumberInput<Double> keyboard = getKeyboard(provider.get().getAcceleration());
              keyboard.show(accelerationInputField,
                  provider.get().getCallbackForAcceleration());
            }
          }
        });
  }

  // Creates a panel for users to input blend radius.
  public <T extends BaseProcessMoveNodeContribution> Component createBlendRadiusPanel(
      final ContributionProvider<T> provider) {

    return createLabelInputField(
        "Blend Radius (m):", blendRadiusInputField, new MouseAdapter() {
          @Override
          public void mousePressed(MouseEvent e) {
            if (blendRadiusInputField.isEditable()) {
              KeyboardNumberInput<Double> keyboard = getKeyboard(provider.get().getBlendRadius());
              keyboard.show(blendRadiusInputField,
                  provider.get().getCallbackForBlendRadius());
            }
          }
        });
  }

  // Creates a GUI for users to input speed, velocity, and acceleration.
  public <T extends BaseProcessMoveNodeContribution> JPanel createParameterInputControls(
      final ContributionProvider<T> provider) {

    JPanel panel = new JPanel();
    panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
    panel.setAlignmentX(Component.LEFT_ALIGNMENT);

    JLabel paramsTitle = new JLabel("Motion Parameters");
    paramsTitle.setFont(paramsTitle.getFont()
        .deriveFont((float) style.getSmallHeaderFontSize())
        .deriveFont(Font.BOLD));
    panel.add(paramsTitle);

    panel.add(createVerticalSpacing());

    panel.add(createSpeedPanel(provider));

    panel.add(createVerticalSpacing());

    panel.add(createAccelerationPanel(provider));

    panel.add(createVerticalSpacing());

    panel.add(createBlendRadiusPanel(provider));

    return panel;
  }
}
