package com.ur.urcap.externalaxis.moveto.toolbar;

import com.ur.urcap.api.contribution.toolbar.ToolbarAPIProvider;
import com.ur.urcap.api.contribution.toolbar.ToolbarContext;
import com.ur.urcap.api.contribution.toolbar.swing.SwingToolbarContribution;
import com.ur.urcap.api.domain.userinteraction.inputvalidation.InputValidationFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputCallback;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardInputFactory;
import com.ur.urcap.api.domain.userinteraction.keyboard.KeyboardTextInput;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptCommand;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptReader;
import com.ur.urcap.externalaxis.moveto.impl.Style;
import com.ur.urcap.externalaxis.moveto.impl.V5Style;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEClient;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEOutputHandler;
import com.ur.urcap.externalaxis.moveto.rtde.RTDEOutputParam;
import com.ur.urcap.externalaxis.moveto.URCommunicator.ScriptSender;

import javax.swing.*;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.stream.Collectors;

class MoveToToolbarContribution implements SwingToolbarContribution {
	private final Style style;
	private final ToolbarAPIProvider apiProvider;
	private final InputValidationFactory validatorFactory;
	private final KeyboardInputFactory keyboardInputFactory;
	private static final int MAXIMUM_FRACTION_DIGITS = 3;
	private static final Integer SOCKET_PORT = 12343;

	// ScriptCommunicator
	private final ScriptSender sender;

	// RTDE
	private final RTDEClient client;

	// DecimalFormat
	private DecimalFormat df = new DecimalFormat("#");

	private final JTextField axisGroupName;
	private final JTextField frameName;
	private final JLabel currentPositionLabel;
	private final JCheckBox frameTrackingCheckBox;
	private final JButton moveAxisButton;
	private final JTextField setAxisGroupPosition;
	private final JSlider speedSlider;
	private final JLabel speedSliderLabel;
	private final JSeparator jSeparator1;
	private final JLabel activeCommandLabel;
	private final JLabel frameLabel;
	private final JLabel commandLabel;

	private final JLabel axisGroupActualPositions = new JLabel("");

	private String receivedString = "";
	private boolean axisGroupValid = false;
	private Integer[] indices = new Integer[0];
	private final JLabel errorLabel = new JLabel("");

	MoveToToolbarContribution(ToolbarContext context) {
		// API and Keyboard
		apiProvider = context.getAPIProvider();
		style = new V5Style();
		keyboardInputFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getKeyboardInputFactory();
		validatorFactory = apiProvider.getUserInterfaceAPI().getUserInteraction().getInputValidationFactory();

		// GUI
		axisGroupName = new JTextField("");
		frameName = new JTextField("");
		setAxisGroupPosition = new JTextField("[]");
		speedSlider = new JSlider();
		speedSliderLabel = new JLabel("Speed [%]");
		frameTrackingCheckBox = new JCheckBox("Frame Tracking");
		currentPositionLabel = new JLabel("Current position:");
		moveAxisButton = new JButton("Move axis to new position");
		jSeparator1 = new JSeparator();
		activeCommandLabel = new JLabel("Last Sent Command(s):");
		frameLabel = new JLabel("Frame:");
		commandLabel = new JLabel("");

		sender = new ScriptSender();
		client = new RTDEClient(new RTDEOutputHandler() {
			@Override
			public void onProtocolVersion(int protocolVersion) {
				if (protocolVersion != 2) {
					throw new RuntimeException("Protocol v2 expected");
				}
			}

			@Override
			public void onData(RTDEOutputParam output, Object value) {
					double[] data = (double[]) value;
					if (axisGroupValid && indices.length > 0) {
						String stringData = Arrays.stream(indices)
								.map(i -> df.format(data[i]))
								.collect(Collectors.joining(", "));
						axisGroupActualPositions.setText("[" + stringData + "]");
					}
				}
			});
		df.setMaximumFractionDigits(MAXIMUM_FRACTION_DIGITS);
	}

	@Override
	public void openView() {
		// Connect to the ScriptSender and RTDEClient
		sender.connect("127.0.0.1", ScriptSender.PRIMARY_PORT, 200);
		if (client.connect("127.0.0.1", RTDEClient.PORT, 200)) {
			try {
				client.requestProtocolVersion(2);
				client.setupOutputsV2(50,
						RTDEOutputParam.actual_external_q);
				client.startOutput();
			} catch (Exception e) {
				System.out.println("RTDE client setup failed");
				System.out.println(e);
			}
		}
	}

	@Override
	public void closeView() {
		sender.disconnect();
		client.disconnect();
	}

	public void buildUI(JPanel jPanel) {
		ChangeListener changeListener = e -> {
			speedSliderLabel.setText("Speed " + speedSlider.getValue() + "%");
		};
		speedSlider.addChangeListener(changeListener);
		speedSlider.setMinimum(1);

		JLabel axisGroupLabel = new JLabel("Axis Group:");
		setupTextInputField(axisGroupName, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForAxisGroupName();
				keyboard.show(axisGroupName, getCallbackForAxisGroupName());
			}
		});

		setupTextInputField(frameName, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForTrackedFrameName();
				keyboard.show(frameName, getCallbackForTrackedFrameName());
			}
		});

		JLabel setAxisGroupPositionLabel = new JLabel("Set Axis Group Position:");
		setupTextInputField(setAxisGroupPosition, new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				KeyboardTextInput keyboard = getKeyboardForAxisGroupPosition();
				keyboard.show(setAxisGroupPosition, getCallbackForAxisGroupPosition());
			}
		});

		frameTrackingCheckBox.setEnabled(false);
		frameTrackingCheckBox.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (!frameTrackingCheckBox.isSelected()) {
					frameName.setText("");
				}
			}
		});


		setupDisplayField(axisGroupActualPositions);

		moveAxisButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent evt) {
				moveAxisButtonMousePressed(evt);
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				moveAxisButtonMouseReleased(e);
			}
		});

		javax.swing.GroupLayout layout = new javax.swing.GroupLayout(jPanel);
		jPanel.setLayout(layout);
		layout.setHorizontalGroup(
				layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
						.addComponent(jSeparator1)
						.addGroup(layout.createSequentialGroup()
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
										.addGroup(layout.createSequentialGroup()
												.addContainerGap()
												.addComponent(commandLabel))
										.addGroup(layout.createSequentialGroup()
												.addGap(237, 237, 237)
												.addComponent(speedSliderLabel)))
								.addGap(0, 0, Short.MAX_VALUE))
						.addGroup(layout.createSequentialGroup()
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
										.addGroup(layout.createSequentialGroup()
												.addGap(6, 6, 6)
												.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
														.addComponent(setAxisGroupPositionLabel)
														.addComponent(currentPositionLabel))
												.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
												.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
														.addGroup(layout.createSequentialGroup()
																.addComponent(frameName, javax.swing.GroupLayout.PREFERRED_SIZE, 120, javax.swing.GroupLayout.PREFERRED_SIZE)
																.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																.addComponent(frameTrackingCheckBox))
														.addComponent(axisGroupActualPositions)
														.addGroup(layout.createSequentialGroup()
																.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
																		.addComponent(setAxisGroupPosition, javax.swing.GroupLayout.DEFAULT_SIZE, 160, Short.MAX_VALUE)
																		.addComponent(axisGroupName))
																.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																.addComponent(errorLabel))))
										.addGroup(layout.createSequentialGroup()
												.addContainerGap()
												.addComponent(activeCommandLabel))
										.addGroup(layout.createSequentialGroup()
												.addGap(6, 6, 6)
												.addComponent(axisGroupLabel))
										.addGroup(layout.createSequentialGroup()
												.addGap(83, 83, 83)
												.addComponent(speedSlider, javax.swing.GroupLayout.PREFERRED_SIZE, 371, javax.swing.GroupLayout.PREFERRED_SIZE))
										.addGroup(layout.createSequentialGroup()
												.addGap(6, 6, 6)
												.addComponent(frameLabel))
										.addGroup(layout.createSequentialGroup()
												.addGap(132, 132, 132)
												.addComponent(moveAxisButton, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)))
								.addContainerGap(86, Short.MAX_VALUE))
		);

		layout.linkSize(javax.swing.SwingConstants.HORIZONTAL, new java.awt.Component[] {axisGroupName, setAxisGroupPosition});

		layout.setVerticalGroup(
				layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
						.addGroup(layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
										.addComponent(axisGroupLabel)
										.addComponent(axisGroupName, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
										.addComponent(errorLabel))
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
										.addComponent(setAxisGroupPositionLabel)
										.addComponent(setAxisGroupPosition, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
										.addComponent(currentPositionLabel)
										.addComponent(axisGroupActualPositions))
								.addGap(18, 18, 18)
								.addComponent(moveAxisButton, javax.swing.GroupLayout.PREFERRED_SIZE, 75, javax.swing.GroupLayout.PREFERRED_SIZE)
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addComponent(speedSlider, javax.swing.GroupLayout.PREFERRED_SIZE, 30, javax.swing.GroupLayout.PREFERRED_SIZE)
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addComponent(speedSliderLabel)
								.addGap(12, 12, 12)
								.addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
										.addComponent(frameLabel)
										.addComponent(frameName, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
										.addComponent(frameTrackingCheckBox))
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
								.addComponent(jSeparator1, javax.swing.GroupLayout.PREFERRED_SIZE, 12, javax.swing.GroupLayout.PREFERRED_SIZE)
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addComponent(activeCommandLabel)
								.addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addComponent(commandLabel)
								.addContainerGap(122, Short.MAX_VALUE))
		);

	}

	private void setupTextInputField(final JTextField inputField, MouseAdapter mouseAdapter) {
		inputField.setFocusable(false);
		inputField.setPreferredSize(style.getInputfieldSize());
		inputField.setMaximumSize(inputField.getPreferredSize());
		inputField.addMouseListener(mouseAdapter);
	}

	private KeyboardTextInput getKeyboardForAxisGroupName() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(axisGroupName.getText());
		keyboardInput.setErrorValidator(validatorFactory.createStringLengthValidator(1, 255));
		return keyboardInput;
	}

	private KeyboardInputCallback<String> getCallbackForAxisGroupName() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				axisGroupName.setText(value);

				ScriptCommand axisIndices = createIndexCommand("/get_axis_indices.script", "getIndices", axisGroupName.getText());

				if (sendCommandAndReceiveResponse(axisIndices)){
					// Parse the received string to get the indices of the axes in the group
					if (receivedString.equals("[]")) {
						errorLabel.setText("Invalid Axis Group");
						errorLabel.setForeground(Color.RED);
						axisGroupValid = false;}
					else {
						String[] splitIndices = receivedString.replaceAll("\\[|\\]", "").split(",");
						indices = Arrays.stream(splitIndices)
								.map(Integer::parseInt)
								.filter(i -> i >= 0)
								.toArray(Integer[]::new);
						axisGroupValid = true;
						errorLabel.setText("");
					}
				}
			}
		};
	}

	private ScriptCommand createIndexCommand(String scriptFileName, String functionName, String axisGroupName) {
		ScriptCommand command = new ScriptCommand(axisGroupName + "_" + functionName );
		command.appendFromFile(scriptFileName);
		command.appendLine("getIndices(\""+ axisGroupName +"\")");
		command.appendLine("if (axisIndices.length()  >= 1):");
		command.appendLine("sort_list(axisIndices)");
		command.appendLine("end");
		command.appendLine("socket_open(\"127.0.0.1\", " + SOCKET_PORT.toString() + ", \"my_socket\")");
		command.appendLine("  socket_send_line(axisIndices, \"my_socket\")");
		command.appendLine("  socket_close(\"my_socket\")");
		return command;
	}

	private boolean sendCommandAndReceiveResponse(ScriptCommand command) {
		// Send the command to the robot and read the result by starting
		// a server socket and reading the data sent by the robot.
		ScriptReader reader = new ScriptReader();
		Thread thread = new Thread(reader);
		thread.start();

		if (sender.executeURScript(command)){
			try {
				thread.join();
			} catch (InterruptedException e) {
			}
			receivedString = reader.getReceivedString();
			return true;
		} else {
			return false;
		}
	}

	private KeyboardTextInput getKeyboardForTrackedFrameName() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(frameName.getText());
		keyboardInput.setErrorValidator(validatorFactory.createStringLengthValidator(1, 255));
		return keyboardInput;
	}

	private KeyboardInputCallback<String> getCallbackForTrackedFrameName() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				frameName.setText(value);
				frameTrackingCheckBox.setEnabled(true);
			}
		};
	}

	private KeyboardTextInput getKeyboardForAxisGroupPosition() {
		KeyboardTextInput keyboardInput = keyboardInputFactory.createStringKeyboardInput();
		keyboardInput.setInitialValue(setAxisGroupPosition.getText());
		return keyboardInput;
	}

	private KeyboardInputCallback getCallbackForAxisGroupPosition() {
		return new KeyboardInputCallback<String>() {
			@Override
			public void onOk(String value) {
				setAxisGroupPosition.setText(value);
			}
		};
	}

	private void setupDisplayField(final JLabel displayLabel) {
		displayLabel.setFocusable(false);
		displayLabel.setPreferredSize(style.getDisplayfieldSize());
		displayLabel.setMaximumSize(displayLabel.getPreferredSize());
	}

	private void moveAxisButtonMousePressed(java.awt.event.MouseEvent evt) {
		// Constructs the primary program to be sent over a socket. The builtin URScript command
		// 'axis_group_movej' is used to move the axis group to the desired position.
		// If the frame tracking checkbox is selected, the robot will track the given frame or axis
		// which was input by the user.
		try {
			ScriptCommand moveToAxisCommand = new ScriptCommand("urcap_external_axis_ethercat_move_to");

			String frameTrackingCommand = String.format(
					"frame_tracking_enable(\"%s\")",
					frameName.getText());

			String moveToAxisCommandString = String.format(
					"axis_group_movej(\"%s\", %s, 0.5, %s)",
					axisGroupName.getText(),
					setAxisGroupPosition.getText(),
					speedSlider.getValue() / 100f
					);

			if (frameTrackingCheckBox.isSelected()) {
				moveToAxisCommand.appendLine(frameTrackingCommand);
			}

			moveToAxisCommand.appendLine(moveToAxisCommandString);

			if (sender.executeURScript(moveToAxisCommand)) {
				String labelText;
				if (frameTrackingCheckBox.isSelected()) {
					labelText = "<html>" + frameTrackingCommand + "<br>" + moveToAxisCommandString + "</html>";
				} else {
					labelText = moveToAxisCommandString;
				}
				commandLabel.setText(labelText);
			} else {
				throw new IOException();
			}
		} catch (IOException e) {
			System.err.println("Failed to construct the primary program.");
			return;
		}
	}
	private void moveAxisButtonMouseReleased(java.awt.event.MouseEvent evt) {
		// Stops the movement of the axis group by calling the builtin URScript command 'axis_group_speedj',
		// and setting the speed to "0.0" for each of the axes in the group.
		ScriptCommand stopCommand = new ScriptCommand("urcap_external_axis_ethercat_stop_move_to");
		String frameTrackingCommand = String.format(
				"frame_tracking_enable(\"%s\")",
				frameName.getText());
		int numOfAxesInGroup = indices.length;
		String zeroCommands ="[" + String.join(", ", Collections.nCopies(numOfAxesInGroup, "0.0")) + "]";

		String axisGroupSpeedJ = "axis_group_speedj(\"" + axisGroupName.getText() + "\", " + zeroCommands + ", 1)";
		if (frameTrackingCheckBox.isSelected()){
			stopCommand.appendLine(String.format(
					"frame_tracking_enable(\"%s\")"
					, frameName.getText()));
		};
		stopCommand.appendLine(axisGroupSpeedJ);
		if (sender.executeURScript(stopCommand)) {
			String labelText;
			if (frameTrackingCheckBox.isSelected()) {
				labelText = "<html>" + frameTrackingCommand + "<br>" + axisGroupSpeedJ + "</html>";
			} else {
				labelText = axisGroupSpeedJ;
			}
			commandLabel.setText(labelText);
		}
	}
}
